import os
import pythoncom
import pywintypes
import win32security
from typing import Any
from wmi import WMI, x_wmi
from pyad import pyadutils
from pyad.aduser import ADUser
from pyad.adquery import ADQuery
from collections import defaultdict
from pyad.adcomputer import ADComputer
from pyad.adcontainer import ADContainer
from pyad.adgroup import ADObject, ADGroup

import ipih

from pih import A
from pih.consts.errors import NotAccesable
from pih.collections import FullName, User
from ActiveDirectoryService.const import *
from pih.tools import if_else, j, n, nn, lw, ne


class ActiveDirectoryApi:
    DEFAULT_AD_QUERY: ADQuery = ADQuery()

    @staticmethod
    def get_printer_list(driver_name: str | None = None) -> list[dict[str, str]]:
        def printer_mapper(item):
            for field in A.CT_FC.PRINTER.ITEM.get_name_list():
                if field in [
                    A.CT_FNC.DESCRIPTION,
                    A.CT_FNC.PORT_NAME,
                ]:
                    item[field] = A.D.get_first_item(item[field])
            return item

        return A.D.map(
            printer_mapper,
            ActiveDirectoryApi.printer_list_filter_by_driver_name(
                ActiveDirectoryApi.get_shared_printer_list(), driver_name
            ),
        )

    @staticmethod
    def printer_list_filter_by_driver_name(
        printer_list: list[dict[str, str]], driver_name: str | None = None
    ) -> list[dict[str, str]]:
        return if_else(
            n(driver_name),
            printer_list,
            lambda: list(
                filter(
                    lambda item: item["driverName"].lower().find(driver_name.lower())
                    != -1,
                    printer_list,
                )
            ),
        )

    @staticmethod
    def get_shared_printer_list() -> list[dict[str, str]]:
        ad_query: ADQuery = ADQuery()
        ad_query.execute_query(
            attributes=A.CT_FC.PRINTER.ITEM.get_name_list(),
            where_clause=(f"objectCategory = 'printqueue'"),
        )
        query_result = [] if ad_query.get_row_count() == 0 else ad_query.get_results()
        result: list[dict[str, str]] = []
        for item in query_result:
            result.append(item)
        return result

    @staticmethod
    def get_computer_description_list(dn: str) -> list[dict[str, Any]]:
        property_container: ADContainer = ADContainer.from_dn(A.CT_AD.PROPERTY_COMPUTER_DN)
        property_list: list[ADGroup] = property_container.get_children(False)
        computer_property_map: dict[str, int] = defaultdict(int)
        property_item: ADGroup
        for property_item in property_list:
            property_member_list: list[ADComputer] = ADGroup.from_dn(
                property_item.dn
            ).get_members()
            ws_property: A.CT_AD.ComputerProperties = A.D.get(
                A.CT_AD.ComputerProperties, property_item.name
            )
            property_member_item: ADObject
            for property_member_item in property_member_list:
                computer_property_map[property_member_item.name] |= ws_property.value
        ad_query: ADQuery = ADQuery()
        name_field_name: str = A.CT_FNC.NAME
        description_field_name: str = A.CT_FNC.DESCRIPTION
        ad_query.execute_query(
            attributes=[name_field_name, description_field_name],
            base_dn=dn,
            where_clause="objectClass = 'computer'",
        )
        query_result: list = (
            [] if ad_query.get_row_count() == 0 else ad_query.get_results()
        )
        result: list[dict[str, Any]] = []
        for query_item in query_result:
            query_item[A.CT_FNC.PROPERTIES] = computer_property_map[
                query_item[name_field_name]
            ]
            query_item[description_field_name] = A.D.get_first_item(
                query_item[description_field_name]
            )
            query_item[A.CT_FNC.DN] = dn
            result.append(query_item)
        return result

    @staticmethod
    def get_user_existance_status_for_child_in_user_home_directory() -> (
        list[UserExistanceStatusItem]
    ):
        return ActiveDirectoryApi.get_user_existance_status_for_child_in_directory(
            A.PTH.USER.HOME_FOLDER
        )

    @staticmethod
    def get_user_login_by_workstation_name_via_wmi(value: str) -> list[str] | None:
        if A.C_R.accessibility_by_ping(value):
            pythoncom.CoInitialize()
            try:
                wmi_object = WMI(value)
                login: str | None = None
                domain: str = A.CT_AD.DOMAIN_NAME.lower()
                for computer in wmi_object.Win32_ComputerSystem():
                    login = computer.userName
                    if ne(login):
                        login = lw(login)
                        domain_index: int = login.find(domain)
                        login = (
                            login[domain_index + len(domain) + 1 :]
                            if domain_index > -1
                            else login
                        )
                return [] if n(login) else [login]  # type: ignore
            except x_wmi:
                raise NotAccesable()
            finally:
                pythoncom.CoUninitialize()
        else:
            raise NotAccesable()

    @staticmethod
    def get_user_login_by_workstation_name_via_query(
        value: str, active: bool | None = True
    ) -> list[str] | None:
        if A.C_R.accessibility_by_ping(value):
            with A.ER.detect():
                return A.EXC.get_users_logged_on(value, active=active)
        else:
            raise NotAccesable()

    @staticmethod
    def get_user_existance_status_for_child_in_directory(
        directory: str,
    ) -> list[UserExistanceStatusItem]:
        dir_list: list[str] = os.listdir(directory)
        result: list[UserExistanceStatusItem] = []
        for dir_item in dir_list:
            login: str = dir_item.lower()
            if login.find(j((".", A.CT_AD.DOMAIN_NAME.lower()))) != -1:
                login = login.split(".")[0]
            status: UserExistanceStatus | None = None
            if ActiveDirectoryApi.user_is_exists_by_login(dir_item, True):
                status = UserExistanceStatus.EXISTS
            elif ActiveDirectoryApi.user_is_exists_by_login(dir_item, False):
                status = UserExistanceStatus.INACTIVE
            else:
                status = UserExistanceStatus.NOT_EXISTS
            result.append(UserExistanceStatusItem(login, status))
        return result

    @staticmethod
    def query_user_by_login_for_dn(login: str, dn: str) -> Any:
        return ActiveDirectoryApi.query_by(A.CT_UP.LOGIN, login, [A.CT_UP.LOGIN], dn)

    @staticmethod
    def query_user_by_name_for_dn(login: str, dn: str) -> Any:
        return ActiveDirectoryApi.query_by(A.CT_UP.NAME, login, [A.CT_UP.NAME], dn)

    @staticmethod
    def user_is_exists_by_login(login: str, active: bool | None = None) -> bool:
        def internal_user_is_exists_by_login(login: str, active: bool) -> bool:
            return nn(
                ActiveDirectoryApi.query_user_by_login_for_dn(
                    login,
                    (
                        A.CT_AD.ACTIVE_USERS_CONTAINER_DN
                        if active
                        else A.CT_AD.INACTIVE_USERS_CONTAINER_DN
                    ),
                )
            )

        if n(active):
            return internal_user_is_exists_by_login(
                login, True
            ) or internal_user_is_exists_by_login(login, False)
        return internal_user_is_exists_by_login(login, active)

    @staticmethod
    def get_user_home_directory_path(login: str) -> str:
        return A.PTH.join(A.PTH.USER.HOME_FOLDER, login)

    @staticmethod
    def get_template_users() -> list[dict]:
        return ActiveDirectoryApi.find_users_by(
            A.CT_UP.LOGIN, A.CT_AD.TEMPLATED_USER_SERACH_TEMPLATE, True
        )

    @staticmethod
    def query_by(
        search_attribute: str,
        search_value: str,
        fields: list[str],
        base_dn: str = A.CT_AD.ACTIVE_USERS_CONTAINER_DN,
        ad_query=DEFAULT_AD_QUERY,
    ) -> dict[str, Any] | None:
        if search_attribute.find("name") != -1:
            if search_value != "*":
                search_value = f"*{search_value}*"
        if search_value == "":
            search_value = "*"
        ad_query.execute_query(
            attributes=fields,
            base_dn=base_dn,
            where_clause=f"objectClass = 'user' and objectCategory = 'person' and {search_attribute} = '{search_value}'",
        )
        if ad_query.get_row_count() == 0:
            return None
        return ad_query.get_results()

    @staticmethod
    def find_by_in_dn(attribute: str, value: str, base_dn: str) -> list[dict[str, Any]]:
        fields: list[str] = A.CT_FC.AD.USER.get_name_list()
        query_result = (
            ActiveDirectoryApi.query_by(attribute, value, fields, base_dn) or []
        )
        result: list[dict] = []
        for query_item in query_result:
            result_item: dict = {}
            for field_item in fields:
                item = query_item[field_item]
                if isinstance(item, (list, tuple)):
                    item = item[0]
                result_item[field_item] = item
            result.append(result_item)
        return result

    @staticmethod
    def user_set_dn(user_dn: str, status: str, container_dn: str) -> None:
        ad_user = ADUser.from_dn(user_dn)
        if status == A.CT_AD.INACTIVE_USERS_CONTAINER_DN:
            ad_user.disable()
        else:
            ad_user.enable()
        try:
            ad_user.move(ADContainer(container_dn))
        except:
            pass

    @staticmethod
    def user_set_password(user_dn: str, password: str) -> None:
        ADUser.from_dn(user_dn).set_password(password)

    @staticmethod
    def user_set_telephone_number(user_dn: str, telephone: str) -> None:
        ADUser.from_dn(user_dn).update_attribute(A.CT_UP.TELEPHONE_NUMBER, telephone)

    @staticmethod
    def user_move(user_dn: str, container_dn: str) -> None:
        user = ADUser.from_dn(user_dn)
        try:
            user.move(ADContainer(container_dn))
        except:
            pass

    @staticmethod
    def get_containers(
        root_dn: str = A.CT_AD.ACTIVE_USERS_CONTAINER_DN,
    ) -> list[dict[str, str]]:
        child_list = ADContainer.from_dn(root_dn).get_children(True)
        result: list = []
        container: ADContainer
        for container in child_list:
            if container.type == "organizationalUnit":
                result.append(
                    {
                        A.CT_UP.NAME: A.D_F.location_list(container.dn, False)[-1],
                        A.CT_UP.DESCRIPTION: A.D.get_first_item(
                            container.get_attribute(A.CT_UP.DESCRIPTION), ""
                        ),
                        A.CT_UP.DN: container.dn,
                    }
                )
        return result

    @staticmethod
    def filter_users(list: list[User], active: bool) -> list[User]:
        return A.D_FL.users_by_dn(
            list,
            (
                A.CT_AD.ACTIVE_USERS_CONTAINER_DN
                if active
                else A.CT_AD.INACTIVE_USERS_CONTAINER_DN
            ),
        )

    @staticmethod
    def find_users_by(attribute: str, value: str, active: bool | None = None) -> list:
        def internal_find_users_by(attribute: str, value: str, active: bool) -> list:
            return ActiveDirectoryApi.find_by_in_dn(
                attribute,
                value,
                (
                    A.CT_AD.ACTIVE_USERS_CONTAINER_DN
                    if active
                    else A.CT_AD.INACTIVE_USERS_CONTAINER_DN
                ),
            )

        if n(active):
            return internal_find_users_by(
                attribute, value, True
            ) + internal_find_users_by(attribute, value, False)
        return internal_find_users_by(attribute, value, active)

    @staticmethod
    def by_group_name_in_container(group_name: str, container_dn: str) -> list[ADGroup]:
        group_name = group_name.lower()
        container: ADContainer = ADContainer.from_dn(container_dn)
        result: list[ADGroup] = []
        container_children = container.get_children(True)
        group_object: ADGroup
        for group_object in container_children:
            name: str = str(group_object.get_attribute(A.CT_UP.NAME)[0]).lower()
            if name == group_name:
                distinguished_name = group_object.dn
                user_list: list[ADUser] = ADGroup.from_dn(
                    distinguished_name
                ).get_members()
                user: ADUser
                for user in user_list:
                    result += ActiveDirectoryApi.find_users_by(A.CT_UP.DN, user.dn)
                break
        return result

    class UserContainer(ADObject):
        def __create_user_object(self, name):
            type: str = "user"
            prefix = "ou" if type == "organizationalUnit" else "cn"
            prefixed_name = j((prefix, name), "=")
            return self._ldap_adsi_obj.Create(type, prefixed_name)

        def create_user(
            self, login, name, password, upn_suffix, optional_attributes={}
        ):
            try:
                upn = j((login, upn_suffix), A.CT.)
                obj = self.__create_user_object(name)
                obj.Put(
                    "sAMAccountName", optional_attributes.get("sAMAccountName", login)
                )
                obj.Put("userPrincipalName", upn)
                obj.SetInfo()
                pyadobj = ADUser.from_com_object(obj)
                pyadobj.enable()
                pyadobj.set_password(password)
                pyadobj.update_attributes(optional_attributes)
                return pyadobj
            except pywintypes.com_error as e:
                pyadutils.pass_up_com_exception(e)

    class ACTION:
        @staticmethod
        def create_user_from_template(
            templated_user_dn: str,
            full_name: FullName,
            login: str,
            password: str,
            description: str,
            telephone: str,
            email: str,
        ) -> bool:
            templated_user = ADUser.from_dn(templated_user_dn)
            group_list = templated_user.get_memberOfs(True)
            result_group_list = {}
            for group in group_list:
                if group.dn not in result_group_list:
                    result_group_list[group.dn] = group
            user_container: ActiveDirectoryApi.UserContainer = (
                ActiveDirectoryApi.UserContainer.from_dn(
                    A.D_Ex.container_dn_from_dn(templated_user_dn)
                )
            )
            name = A.D.fullname_to_string(full_name)
            new_user: ADUser | None = user_container.create_user(
                login,
                name,
                password,
                A.CT_AD.DOMAIN_MAIN,
                optional_attributes={
                    "givenName": A.D.to_given_name(full_name),
                    "sn": full_name.last_name,
                    "userPrincipalName": A.D_F.user_principal_name(login),
                    "homeDirectory": A.PTH.for_windows(
                        j((A.PTH_U.HOME_FOLDER_FULL, login), "\\")
                    ),
                    "displayName": name,
                    "homeDrive": A.CT_AD.USER_HOME_FOLDER_DISK,
                    "mail": email,
                    "description": description,
                    "telephonenumber": telephone,
                },
            )
            new_user.set_user_account_control_setting("PASSWD_NOTREQD", False)
            new_user.set_user_account_control_setting("DONT_EXPIRE_PASSWD", True)
            for item in result_group_list:
                new_user.add_to_group(result_group_list[item])

            return True

        @staticmethod
        def authenticate(login: str, password: str) -> bool:
            try:
                win32security.LogonUser(
                    login,
                    A.CT_AD.DOMAIN_NAME,
                    password,
                    win32security.LOGON32_LOGON_NETWORK,
                    win32security.LOGON32_PROVIDER_DEFAULT,
                )
                return True
            except win32security.error as _:
                return False

        @staticmethod
        def user_remove(user_dn: str) -> bool:
            try:
                ADUser.from_dn(user_dn).delete()
                return True
            except:
                return False
        @staticmethod
        def user_set_telephone_number(user_dn: str, telephone: str) -> bool:
            ActiveDirectoryApi.user_set_telephone_number(user_dn, telephone)
            return True

        @staticmethod
        def user_set_password(user_dn: str, password: str) -> bool:
            ActiveDirectoryApi.user_set_password(user_dn, password)
            return True

        @staticmethod
        def user_move(user_dn: str, container_dn: str) -> bool:
            ActiveDirectoryApi.user_move(user_dn, container_dn)
            return True

        @staticmethod
        def user_set_status(user_dn: str, status: str, container_dn: str) -> bool:
            ActiveDirectoryApi.user_set_dn(user_dn, status, container_dn)
            return True


import ipih

from pih import A, PIHThreadPoolExecutor, thread
from ActiveDirectoryService.const import SD


ISOLATED = False

SC = A.CT_SC

from typing import Any
from pih.collections import FullName, Result
from pih.tools import ParameterList, e, n, ne, while_excepted, one


class DH:
    computer_cache: dict[str, dict[str, Any]] = {}
    computer_description_list: list[dict[str, Any]] = []
    computer_lock: bool = False
    user_cache: dict[str, dict[str, Any]] = {}
    user_lock: bool = False


def start(as_standalone: bool = False) -> None:

    if A.U.for_service(SD, as_standalone=as_standalone):
       
        from pih.consts.errors import NotAccesable
        from ActiveDirectoryService.api import ActiveDirectoryApi as Api
        
        from concurrent.futures import as_completed
        import pythoncom
        
        def service_call_handler(sc: SC, pl: ParameterList) -> Any:
            with A.ER.detect(pythoncom.CoUninitialize):
                pythoncom.CoInitialize()
                if sc == SC.heart_beat:
                    if (
                        A.SE.life_time.seconds % A.CT.CACHE.TTL.WORKSTATIONS
                    ) == 0:
                        update_computer_cache()
                        return
                    if (A.SE.life_time.seconds % A.CT.CACHE.TTL.USERS) == 0:
                        update_users_cache()
                        return
                if sc == SC.drop_user_cache:
                    update_users_cache()
                    return
                if sc == SC.drop_workstaion_cache:
                    update_computer_cache()
                    return
                if sc == SC.check_user_exists_by_login:
                    return Api.user_is_exists_by_login(pl.next(), pl.next())
                if sc == SC.get_user_by_full_name:
                    return get_user_by_attribute(
                        A.CT_FNC.SEARCH_ATTRIBUTE_NAME,
                        A.D.fullname_to_string(pl.next(FullName())),
                        pl.next(),
                    )
                if sc == SC.get_users_by_name:
                    value: str = pl.next()
                    value = value.lower()
                    active: bool | None = pl.next()
                    cached: bool | None = pl.next()
                    cached = A.D.check_not_none(cached, cached, A.S_U.use_cache())
                    if cached:
                        return create_user_result(
                            A.D.filter(
                                lambda item: isinstance(item, dict)
                                and item[A.CT_FNC.SEARCH_ATTRIBUTE_NAME]
                                .lower()
                                .find(value)
                                != -1,
                                A.D.as_list(DH.user_cache),
                            )
                        )
                    return get_user_by_attribute(
                        A.CT_FNC.SEARCH_ATTRIBUTE_NAME, value, active
                    )
                if sc == SC.get_user_by_login:
                    value: str = pl.next()
                    value = value.lower()
                    active: bool | None = pl.next()
                    cached: bool | None = pl.next()
                    cached = A.D.check_not_none(cached, cached, A.S_U.use_cache())
                    if not cached or value not in DH.user_cache:
                        fetch_user_by_login(value, active)
                    return create_user_result(DH.user_cache[value])
                if sc == SC.get_user_by_telephone_number:
                    return get_user_by_attribute(
                        A.CT_FNC.TELEPHONE_NUMBER,
                        pl.next(),
                        pl.next(),
                    )
                if sc == SC.get_template_users:
                    return Result(
                        A.CT_FC.AD.TEMPLATED_USER,
                        Api.get_template_users(),
                    )
                if sc == SC.create_user_by_template:
                    templated_user_dn: str = pl.next()
                    full_name: FullName = pl.next(FullName())
                    login: str = pl.next()
                    password: str = pl.next()
                    description: str = pl.next()
                    telephone: str = pl.next()
                    email: str = pl.next()
                    result: bool = Api.ACTION.create_user_from_template(
                        templated_user_dn,
                        full_name,
                        login,
                        password,
                        description,
                        telephone,
                        email,
                    )
                    update_users_cache()
                    return result
                if sc == SC.set_user_telephone_number:
                    return Api.ACTION.user_set_telephone_number(pl.next(), pl.next())
                if sc == SC.authenticate:
                    return Api.ACTION.authenticate(pl.next(), pl.next())
                if sc == SC.set_user_password:
                    return Api.ACTION.user_set_password(pl.next(), pl.next())
                if sc == SC.set_user_status:
                    return Api.ACTION.user_set_status(
                        pl.next(),
                        pl.next(),
                        pl.next(),
                    )
                if sc == SC.get_containers:
                    return Result(A.CT_FC.AD.CONTAINER, Api.get_containers())
                if sc == SC.get_printer_list:
                    return Result(A.CT_FC.PRINTER.ITEM, Api.get_printer_list())
                target_sc_list: list[SC] = [
                    SC.get_user_list_by_job_position,
                    SC.get_user_list_by_group,
                    SC.get_user_list_by_property,
                ]
                sc_index: int = target_sc_list.index(sc) if sc in target_sc_list else -1
                if sc_index != -1:
                    value: str = pl.next()
                    return create_user_result(
                        while_excepted(
                            lambda: Api.by_group_name_in_container(
                                value,
                                [
                                    A.CT_AD.JOB_POSITION_CONTAINER_DN,
                                    A.CT_AD.GROUP_CONTAINER_DN,
                                    A.CT_AD.PROPERTY_USER_DN,
                                ][sc_index],
                            )
                        )
                    )
                if sc == SC.remove_user:
                    return Api.ACTION.user_remove(pl.next())
                if sc == SC.get_computer_list:
                    container_dn: str = pl.next()
                    return A.R.pack(
                        (
                            A.CT_FCA.WORKSTATION
                            if container_dn == A.CT_AD.WORKSTATIONS_CONTAINER_DN
                            else A.CT_FCA.SERVER
                        ),
                        get_computer_list_by_container_dn(container_dn),
                    )
                if sc == SC.get_computer_description_list:
                    return A.R.pack(
                        A.CT_FCA.COMPUTER_DESCRIPTION,
                        get_computer_discription_list_by_container_dn(pl.next()),
                    )
                if sc == SC.get_workstation_list_by_user_login:
                    login: str = pl.next()
                    login = login.lower()
                    return A.R.pack(
                        A.CT_FCA.WORKSTATION,
                        A.D.filter(
                            lambda item: (
                                login in item[A.CT_FNC.ACTIVE_USERS_LOGIN]
                                if A.D.if_is_in(item, A.CT_FNC.ACTIVE_USERS_LOGIN)
                                else False
                            ),
                            get_computer_list_by_container_dn(
                                A.CT_AD.WORKSTATIONS_CONTAINER_DN
                            ),
                        ),
                    )
                if sc == SC.get_user_by_workstation:
                    name: str = pl.next()
                    return A.R.pack(
                        A.CT_FC.AD.USER,
                        A.D.filter(
                            lambda item: str(item[A.CT_FNC.NAME]).lower() == name,
                            get_computer_list_by_container_dn(),
                        ),
                    )

        def get_computer_list_by_container_dn(value: str | None = None) -> dict:
            computer_cache: dict[str, dict[str, Any]] = DH.computer_cache
            if e(computer_cache):
                update_computer_cache()
            return A.D.filter(
                lambda item: True if n(value) else item[A.CT_FNC.DN] == value,
                [computer_cache[computer_name] for computer_name in computer_cache],
            ) # type: ignore

        def get_computer_discription_list_by_container_dn(
            value: str | None = None,
        ) -> dict[str, dict[str, Any]]:
            computer_description_list: list[dict[str, Any]] = (
                DH.computer_description_list
            )
            if e(computer_description_list):
                update_computer_cache()
            return A.D.filter(
                lambda item: True if n(value) else item[A.CT_FNC.DN] == value,
                computer_description_list,
            ) # type: ignore

        def create_user_result(data: Any) -> Result:
            return Result(A.CT_FC.AD.USER, data)

        def update_computer_cache() -> None:
            if not DH.computer_lock:
                DH.computer_lock = True
                with A.ER.detect():
                    DH.computer_description_list = (
                        Api.get_computer_description_list(
                            A.CT_AD.WORKSTATIONS_CONTAINER_DN
                        )
                        + Api.get_computer_description_list(
                            A.CT_AD.SERVERS_CONTAINER_DN
                        )
                    )
                    computer_cache: dict[str, dict[str, Any]] = {}
                    with PIHThreadPoolExecutor(
                        max_workers=len(DH.computer_description_list)
                    ) as executor:

                        future_to_computer = {
                            executor.submit(
                                Api.get_user_login_by_workstation_name_via_query,
                                value=computer_description_item[A.CT_FNC.NAME],
                                active=True,
                            ): computer_description_item
                            for computer_description_item in DH.computer_description_list
                        }
                        for future_computer in as_completed(future_to_computer): # type: ignore
                            computer: dict[str, Any] = future_to_computer[
                                future_computer
                            ] # type: ignore
                            computer_name: str = computer[A.CT_FNC.NAME].lower()
                            computer_cache[computer_name] = computer
                            exception: BaseException | None = (
                                future_computer.exception()
                            )
                            if e(exception):
                                computer[A.CT_FNC.ACCESSABLE] = True
                                login_list: list[str] = future_computer.result()
                                computer[A.CT_FNC.ACTIVE_USERS_LOGIN] = login_list
                                if ne(login_list):
                                    computer[A.CT_FNC.LOGIN] = login_list[0]
                            else:
                                if isinstance(exception, NotAccesable):
                                    computer[A.CT_FNC.ACCESSABLE] = False
                                    computer[A.CT_FNC.ACTIVE_USERS_LOGIN] = None
                                else:
                                    raise exception # type: ignore
                DH.computer_cache = computer_cache
                DH.computer_lock = False

        def get_user_by_attribute(
            attribute: str,
            value: str,
            active: bool | None = None,
            result_return: bool = True,
        ) -> (
            Result[list[dict[str, Any]] | dict[str, Any]]
            | list[dict[str, Any]]
            | dict[str, Any]
        ):
            data: dict[str, Any] = while_excepted(
                lambda: Api.find_users_by(attribute, value, active)
            )
            return create_user_result(data) if result_return else data

        def fetch_user_by_login(value: str | None, active: bool | None = None) -> None:
            if not DH.user_lock:
                with A.ER.detect():
                    DH.user_lock = True

                    class LDH:
                        user_cache: dict[str, dict[str, Any]] = {}

                    while True:
                        with A.ER.detect():
                            result: list[dict[str, Any]] = get_user_by_attribute(
                                A.CT_FNC.SEARCH_ATTRIBUTE_LOGIN,
                                value or A.CT_AD.SEARCH_ALL_PATTERN,
                                active,
                                False,
                            ) # type: ignore

                            def every_action(user_data: dict[str, Any]) -> None:
                                LDH.user_cache |= {
                                    user_data[A.CT_UP.LOGIN].lower(): user_data
                                }

                            A.D.every(every_action, result) # type: ignore
                            break
                    if n(value):
                        DH.user_cache = LDH.user_cache
                    else:
                        DH.user_cache[value] = (lambda item: item, one)[ # type: ignore
                            len(LDH.user_cache) == 1
                        ](A.D.as_list(LDH.user_cache))
                DH.user_lock = False

        def update_users_cache() -> None:
            fetch_user_by_login(None)

        def service_starts_handler() -> None:
            A.SRV_A.subscribe_on(SC.heart_beat, name="User computer update")
            DH.computer_lock = False
            DH.user_lock = False

            def start_action() -> None:
                pythoncom.CoInitialize()
                update_users_cache()
                update_computer_cache()
                pythoncom.CoUninitialize()

            thread(start_action) # type: ignore

        A.SRV_A.serve(
            SD,
            service_call_handler,
            service_starts_handler, # type: ignore
            isolate=ISOLATED,
            as_standalone=as_standalone,
        )


if __name__ == "__main__":
    start()



import ipih

from enum import Enum
from dataclasses import dataclass
from pih.consts.hosts import Hosts
from pih.collections.service import ServiceDescription


NAME: str = "ActiveDirectory"

HOST = Hosts.DC2

VERSION: str = "0.14"

PACKAGES: tuple[str, ...] = ("pyad", "pywin32", "wmi")

SD: ServiceDescription = ServiceDescription(
    name=NAME,
    description="Active directory service",
    host=HOST.NAME,
    packages=PACKAGES,
    commands=(
        "authenticate",
        "check_user_exists_by_login",
        "get_user_by_full_name",
        "get_users_by_name",
        "get_user_by_login",
        "get_user_by_telephone_number",
        "get_template_users",
        "get_containers",
        "get_users_by_job_position",
        "get_users_by_group",
        "get_printers",
        "get_computer_description_list",
        "get_computer_list",
        "get_workstation_list_by_user_login",
        "get_user_by_workstation",
        "create_user_by_template",
        "create_user_in_container",
        "set_user_telephone_number",
        "set_user_password",
        "set_user_status",
        "remove_user",
        "drop_user_cache",
        "drop_workstaion_cache",
        "get_user_list_by_property",
    ),
    version=VERSION,
    standalone_name="ad",
    use_standalone=True,
)

ANY_SYMBOL: str = "*"


class UserExistanceStatus(Enum):
    EXISTS = 1
    INACTIVE = -1
    NOT_EXISTS = 0


@dataclass
class UserExistanceStatusItem:
    login: str | None = None
    status: UserExistanceStatus | None = None
